from .utils import eprint


def _server_getitem(self, name):
    session_id = self._get_session_id()
    self._request(session_id, name=name)
    response = self._recv_response(session_id)
    if response["success"]:
        return response["data"]["value"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_getitem(self, session_id, request):
    try:
        value = self._parent[request["data"]["name"]]
        self._respond_ok(session_id, value=value, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_setitem(self, name, value):
    session_id = self._get_session_id()
    self._request(session_id, name=name, value=value)
    response = self._recv_response(session_id)

    if not response["success"]:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_setitem(self, session_id, request):
    try:
        data = request["data"]
        self._parent[data["name"]] = data["value"]
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_delitem(self, name):
    session_id = self._get_session_id()
    self._request(session_id, name=name)
    response = self._recv_response(session_id)

    if not response["success"]:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_delitem(self, session_id, request):
    try:
        del self._parent[request["data"]["name"]]
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_pop(self, name):
    session_id = self._get_session_id()
    self._request(session_id, name=name)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["value"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_pop(self, session_id, request):
    try:
        value = self._parent.pop(request["data"]["name"])
        self._respond_ok(session_id, value=value, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_iter(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["iter"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_iter(self, session_id, request):
    try:
        self._respond_ok(
            session_id, iter=iter(self._parent), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_contains(self, name):
    session_id = self._get_session_id()
    self._request(session_id, name=name)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["contains"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_contains(self, session_id, request):
    try:
        contains = (request["data"]["name"] in self._parent)
        self._respond_ok(
            session_id, contains=contains, last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_len(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["len"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_len(self, session_id, request):
    try:
        self._respond_ok(session_id, len=self._parent.__len__(), last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_keys(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["keys"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_keys(self, session_id, request):
    try:
        self._respond_ok(session_id, keys=self._parent.keys(), last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_values(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["values"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_values(self, session_id, request):
    try:
        self._respond_ok(
            session_id, values=self._parent.values(), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_items(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["items"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_items(self, session_id, request):
    try:
        self._respond_ok(session_id, items=self._parent.items(), last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_clear(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if not response["success"]:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_clear(self, session_id, request):
    try:
        self._parent.clear()
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def init_global_dict_methods(cls):
    cls._server_getitem = _server_getitem
    cls._server_setitem = _server_setitem
    cls._server_delitem = _server_delitem
    cls._server_pop = _server_pop
    cls._server_iter = _server_iter
    cls._server_contains = _server_contains
    cls._server_len = _server_len
    cls._server_keys = _server_keys
    cls._server_values = _server_values
    cls._server_items = _server_items
    cls._server_clear = _server_clear
    cls._process__server_getitem = _process__server_getitem
    cls._process__server_setitem = _process__server_setitem
    cls._process__server_delitem = _process__server_delitem
    cls._process__server_clear = _process__server_clear
    cls._process__server_keys = _process__server_keys
    cls._process__server_values = _process__server_values
    cls._process__server_items = _process__server_items
    cls._process__server_iter = _process__server_iter
    cls._process__server_contains = _process__server_contains
    cls._process__server_len = _process__server_len
    cls._process__server_pop = _process__server_pop

    return cls



import threading
import os

from .utils import file_size, md5, eprint
from .Config import Config
from .NodeInternalClasses import Future, Thread


def _write_to_file(self, data, file_name):
    self._request(self._get_session_id(), data=data, file_name=file_name)


def _process__write_to_file(self, request):
    try:
        file_name = os.path.abspath(request["data"]["file_name"])
        if self._out_file is not None and self._out_file.name != file_name:
            self._out_file.close()
            self._out_file = None

        if self._out_file is None:
            folder_path = os.path.dirname(file_name)
            if not os.path.isdir(folder_path):
                os.makedirs(folder_path)

            self._out_file = open(file_name, "wb")

        self._out_file.write(request["data"]["data"])
        self._out_file.flush()
    except BaseException:
        pass


def _close_file(self):
    self._request(self._get_session_id())


def _process__close_file(self, request):
    try:
        if self._out_file is not None:
            self._out_file.close()
            self._out_file = None
    except BaseException:
        pass


def get_file(self, src_file_name, dest_file_name=None, block=True):
    session_id = self._get_session_id()

    if dest_file_name is None:
        dest_file_name = os.path.basename(src_file_name)

    self._request(
        session_id, src_file_name=src_file_name,
        file_size=file_size(dest_file_name),
        md5=md5(dest_file_name), block=block
    )

    def session():
        response = self._recv_response(session_id)
        if response["cancel"] or response["data"]["same_file"]:
            return

        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]

        folder_path = os.path.dirname(os.path.abspath(dest_file_name))
        if not os.path.isdir(folder_path):
            os.makedirs(folder_path)

        try:
            file = open(dest_file_name, "wb")
            self._respond_ok(session_id)
        except BaseException as e:
            self._respond_exception(session_id, e)
            if block:
                raise e
            else:
                return

        recved_size = 0
        full_size = response["data"]["file_size"]
        while recved_size < full_size:
            response = self._recv_response(session_id)
            if response["cancel"]:
                break
            data = response["data"]["data"]
            file.write(data)
            file.flush()
            recved_size += len(data)
        file.close()

    if block:
        session()
    else:
        thread = threading.Thread(target=session)
        thread.start()
        return Future(self, session_id)


def _process_get_file(self, session_id, request):
    def session():
        data = request["data"]
        src_file_name = data["src_file_name"]
        src_file_size = file_size(src_file_name)

        same_file = (src_file_size != 0 and
                     data["file_size"] == src_file_size and
                     data["md5"] == md5(src_file_name))
        try:
            file = open(src_file_name, "rb")
            self._respond_ok(
                session_id, same_file=same_file,
                file_size=src_file_size, last_one=same_file
            )
        except BaseException as e:
            self._respond_exception(
                session_id, e, same_file=same_file, block=request["block"]
            )
            self._make_signal(session_id)
            return

        if same_file:
            if not request["block"]:
                self._put_result(session_id)
            self._make_signal(session_id)
            return

        response = self._recv_response(session_id)
        if not response["success"]:
            if not request["block"]:
                self._put_exception(session_id, response["exception"])
            self._make_signal(session_id)
            return

        sent_size = 0
        while sent_size < src_file_size:
            data = file.read(self.send_buffer)
            sent_size += len(data)
            self._respond_ok(
                session_id, data=data,
                last_one=(sent_size == src_file_size)
            )
        file.close()

        if not request["block"]:
            self._put_result(session_id)

        self._make_signal(session_id)

    thread = Thread(target=session)
    thread.start()
    signal = self._recv_signal(session_id)
    if signal["cancel"] and thread.is_alive():
        thread.kill()
        thread.join()
        self._respond_ok(session_id, cancel=True, last_one=True)
        self._put_result(session_id, cancelled=True)
    else:
        thread.join()


def put_file(self, src_file_name, dest_file_name=None, block=True):
    file = open(src_file_name, "rb")
    session_id = self._get_session_id()

    if dest_file_name is None:
        dest_file_name = os.path.basename(src_file_name)

    src_file_size = file_size(src_file_name)

    self._stop_send = False
    self._request(
        session_id, md5=md5(src_file_name),
        file_size=src_file_size,
        dest_file_name=dest_file_name,
        block=block
    )

    def session():
        response = self._recv_response(session_id)
        if response["cancel"] or response["data"]["same_file"]:
            return

        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]

        sent_size = 0
        while sent_size < src_file_size:
            if self._stop_send:
                break
            data = file.read(self.send_buffer)
            sent_size += len(data)
            self._respond_ok(
                session_id, data=data,
                last_one=(sent_size == src_file_size)
            )
        file.close()

    if block:
        session()
    else:
        thread = Thread(target=session)
        thread.start()
        return Future(self, session_id)


def _process_put_file(self, session_id, request):
    def session():
        data = request["data"]
        dest_file_name = data["dest_file_name"]
        full_size = data["file_size"]

        if full_size == file_size(dest_file_name) and \
           data["md5"] == md5(dest_file_name):
            self._respond_ok(
                session_id, same_file=True,
                block=request["block"], last_one=True
            )
            self._make_signal(session_id)
            return

        folder_path = os.path.dirname(os.path.abspath(dest_file_name))
        if not os.path.isdir(folder_path):
            os.makedirs(folder_path)

        try:
            file = open(dest_file_name, "wb")
            self._respond_ok(session_id, same_file=False)
        except BaseException as e:
            self._respond_exception(
                session_id, e, same_file=False, block=request["block"]
            )
            self._make_signal(session_id)
            return

        recved_size = 0
        while recved_size < full_size:
            response = self._recv_response(session_id)
            if response["cancel"]:
                break
            data = response["data"]["data"]
            file.write(data)
            file.flush()
            recved_size += len(data)
        file.close()

        if not request["block"]:
            self._put_result(session_id)

        self._make_signal(session_id)

    thread = Thread(target=session)
    thread.start()
    signal = self._recv_signal(session_id)
    if signal["cancel"] and thread.is_alive():
        thread.kill()
        thread.join()
        self._respond_ok(session_id, cancel=True, last_one=True)
        self._put_result(session_id, cancelled=True)
    else:
        thread.join()


def get_folder(self, src_folder_name, dest_folder_name=None, block=True):
    session_id = self._get_session_id()
    if dest_folder_name is None:
        dest_folder_name = os.path.basename(src_folder_name)

    if Config.debug:
        self._request(session_id, src_folder_name=src_folder_name, block=block,
                      debug=f"I need folder {src_folder_name}")
    else:
        self._request(session_id, src_folder_name=src_folder_name, block=block)

    def session():
        response = self._recv_response(session_id)
        if response["cancel"]:
            return

        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]

        for folder in response["data"]["folders"]:
            folder_path = dest_folder_name + "/" + folder
            if not os.path.isdir(folder_path):
                try:
                    os.makedirs(folder_path)
                except BaseException:
                    eprint(self._traceback(False))

        if Config.debug:
            self._respond_ok(session_id,
                             debug="Created all folders. Please send files.")
        else:
            self._respond_ok(session_id)

        while True:
            response = self._recv_response(session_id)
            data = response["data"]
            if response["cancel"] or data["finished"]:
                return

            if not response["success"]:
                eprint(response["traceback"])
                continue

            dest_file_name = dest_folder_name + "/" + data["file_name"]
            same_file = (data["file_size"] == file_size(dest_file_name) and
                         data["md5"] == md5(dest_file_name))

            if same_file:
                if Config.debug:
                    self._respond_ok(
                        session_id, same_file=same_file,
                        debug=f"Already have file: {dest_file_name}"
                    )
                else:
                    self._respond_ok(session_id, same_file=same_file)

                continue
            else:
                try:
                    file = open(dest_file_name, "wb")
                    if Config.debug:
                        self._respond_ok(
                            session_id, same_file=same_file,
                            debug="I don't have this file, please send me."
                        )
                    else:
                        self._respond_ok(session_id, same_file=same_file)
                except BaseException as e:
                    eprint(self._traceback())
                    if Config.debug:
                        self._respond_exception(session_id, e,
                                                debug="Error")
                    else:
                        self._respond_exception(session_id, e)
                    continue

            recved_size = 0
            full_size = response["data"]["file_size"]
            while recved_size < full_size:
                response = self._recv_response(session_id)
                if response["cancel"]:
                    return
                data = response["data"]["data"]
                file.write(data)
                file.flush()
                recved_size += len(data)
            file.close()

    if block:
        session()
    else:
        thread = threading.Thread(target=session)
        thread.start()
        return Future(self, session_id)


def _process_get_folder(self, session_id, request):
    def session():
        src_folder_name = request["data"]["src_folder_name"]
        if not os.path.isdir(src_folder_name):
            e = FileNotFoundError(f"{src_folder_name} is not a folder.")
            self._respond_exception(session_id, e, block=request["block"])
            self._make_signal(session_id)
            return
        i = len(src_folder_name)-1
        while src_folder_name[i] in ['/', '\\']:
            i -= 1

        folders = []
        for root, dirs, files in os.walk(src_folder_name):
            for name in dirs:
                folders.append(os.path.join(root, name)[i+2:])
        if Config.debug:
            self._respond_ok(session_id, folders=folders,
                             debug=f"Please create folders: {folders}")
        else:
            self._respond_ok(session_id, folders=folders)

        response = self._recv_response(session_id)

        for root, dirs, files in os.walk(src_folder_name):
            for name in files:
                src_file_name = os.path.join(root, name)
                src_file_size = file_size(src_file_name)
                file_name = src_file_name[i+2:]

                try:
                    file = open(src_file_name, "rb")
                    if Config.debug:
                        self._respond_ok(
                            session_id, finished=False,
                            file_name=file_name,
                            file_size=src_file_size, md5=md5(src_file_name),
                            debug=f"Do you have file {src_file_name} ?"
                        )
                    else:
                        self._respond_ok(
                            session_id, finished=False,
                            file_name=file_name,
                            file_size=src_file_size, md5=md5(src_file_name)
                        )
                except BaseException as e:
                    if Config.debug:
                        self._respond_exception(
                            session_id, e,
                            debug=f"I cannot open file: {src_file_name}"
                        )
                    else:
                        self._respond_exception(session_id, e)
                    continue

                response = self._recv_response(session_id)
                if not response["success"] or response["data"]["same_file"]:
                    continue

                sent_size = 0
                while sent_size < src_file_size:
                    data = file.read(self.send_buffer)
                    if Config.debug:
                        message = f"Please write {len(data)} bytes data to file {file_name}"
                        self._respond_ok(
                            session_id, data=data,
                            debug=message
                        )
                    else:
                        self._respond_ok(session_id, data=data)
                    sent_size += len(data)
                file.close()

        if Config.debug:
            self._respond_ok(session_id, finished=True, last_one=True,
                             debug="Sent all files.")
        else:
            self._respond_ok(session_id, finished=True, last_one=True)

        if not request["block"]:
            self._put_result(session_id)

        self._make_signal(session_id)

    thread = Thread(target=session)
    thread.start()
    signal = self._recv_signal(session_id)
    if signal["cancel"] and thread.is_alive():
        thread.kill()
        thread.join()
        if Config.debug:
            self._respond_ok(session_id, cancel=True, last_one=True,
                             debug="Cancel.")
        else:
            self._respond_ok(session_id, cancel=True, last_one=True)

        self._put_result(session_id, cancelled=True)
    else:
        thread.join()


def put_folder(self, src_folder_name, dest_folder_name=None, block=True):
    if not os.path.isdir(src_folder_name):
        raise FileNotFoundError(src_folder_name + " is not a folder.")

    session_id = self._get_session_id()

    if dest_folder_name is None:
        dest_folder_name = os.path.basename(src_folder_name)

    self._stop_send = False
    if Config.debug:
        self._request(
            session_id, dest_folder_name=dest_folder_name, block=block,
            debug=f"I will send you folder: {dest_folder_name}"
        )
    else:
        self._request(
            session_id, dest_folder_name=dest_folder_name, block=block
        )

    def session():
        response = self._recv_response(session_id)
        if response["cancel"]:
            return

        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]

        i = len(src_folder_name)-1
        while src_folder_name[i] in ['/', '\\']:
            i -= 1

        folders = []
        for root, dirs, files in os.walk(src_folder_name):
            for name in dirs:
                folders.append(os.path.join(root, name)[i+2:])
        if Config.debug:
            self._respond_ok(session_id, folders=folders,
                             debug=f"Please create folders: {folders}")
        else:
            self._respond_ok(session_id, folders=folders)

        response = self._recv_response(session_id)
        if response["cancel"]:
            return

        for root, dirs, files in os.walk(src_folder_name):
            for name in files:
                src_file_name = os.path.join(root, name)
                src_file_size = file_size(src_file_name)
                file_name = src_file_name[i+2:]

                try:
                    file = open(src_file_name, "rb")
                    if Config.debug:
                        self._respond_ok(
                            session_id, finished=False,
                            file_name=file_name,
                            file_size=src_file_size, md5=md5(src_file_name),
                            debug=f"Do you have file {file_name} ?"
                        )
                    else:
                        self._respond_ok(
                            session_id, finished=False,
                            file_name=file_name,
                            file_size=src_file_size, md5=md5(src_file_name)
                        )

                except BaseException as e:
                    if Config.debug:
                        self._respond_exception(session_id, e,
                                                debug="Error")
                    else:
                        self._respond_exception(session_id, e)

                    continue

                response = self._recv_response(session_id)
                if response["cancel"]:
                    return

                if not response["success"] or response["data"]["same_file"]:
                    continue

                sent_size = 0
                while sent_size < src_file_size:
                    if self._stop_send:
                        return
                    data = file.read(self.send_buffer)
                    if Config.debug:
                        message = f"Please write {len(data)} bytes to file {file_name}"
                        self._respond_ok(
                            session_id, data=data,
                            debug=message
                        )
                    else:
                        self._respond_ok(session_id, data=data)

                    sent_size += len(data)
                file.close()

        if Config.debug:
            self._respond_ok(
                session_id, finished=True, last_one=True,
                debug="Sent all files."
            )
        else:
            self._respond_ok(session_id, finished=True, last_one=True)

    if block:
        session()
    else:
        thread = threading.Thread(target=session)
        thread.start()
        return Future(self, session_id)


def _process_put_folder(self, session_id, request):
    def session():
        dest_folder_name = request["data"]["dest_folder_name"]
        try:
            if not os.path.isdir(dest_folder_name):
                os.makedirs(dest_folder_name)

            if Config.debug:
                self._respond_ok(session_id,
                                 debug=f"Created folder {dest_folder_name}")
            else:
                self._respond_ok(session_id)

        except BaseException as e:
            if Config.debug:
                self._respond_exception(
                    session_id, e, block=request["block"],
                    debug=f"Cannot create folder {dest_folder_name}"
                )
            else:
                self._respond_exception(session_id, e, block=request["block"])

            self._make_signal(session_id)
            return

        response = self._recv_response(session_id)
        for folder in response["data"]["folders"]:
            folder_path = dest_folder_name + "/" + folder
            if not os.path.isdir(folder_path):
                try:
                    os.makedirs(folder_path)
                except BaseException:
                    pass

        if Config.debug:
            self._respond_ok(session_id,
                             debug="Created all folders.")
        else:
            self._respond_ok(session_id)

        while True:
            response = self._recv_response(session_id)
            if not response["success"]:
                continue

            data = response["data"]
            if data["finished"]:
                break

            dest_file_name = dest_folder_name + "/" + data["file_name"]
            same_file = (data["file_size"] == file_size(dest_file_name) and
                         data["md5"] == md5(dest_file_name))

            if same_file:
                if Config.debug:
                    self._respond_ok(
                        session_id, same_file=same_file,
                        debug=f"Already have file {dest_file_name}"
                    )
                else:
                    self._respond_ok(session_id, same_file=same_file)

                continue
            else:
                try:
                    file = open(dest_file_name, "wb")
                    if Config.debug:
                        self._respond_ok(
                            session_id, same_file=same_file,
                            debug="I don't have this file, please send me."
                        )
                    else:
                        self._respond_ok(session_id, same_file=same_file)
                except BaseException as e:
                    if Config.debug:
                        self._respond_exception(session_id, e,
                                                debug="Error")
                    else:
                        self._respond_exception(session_id, e)

                    continue

            recved_size = 0
            full_size = response["data"]["file_size"]
            while recved_size < full_size:
                response = self._recv_response(session_id)
                if response["cancel"]:
                    break
                data = response["data"]["data"]
                file.write(data)
                file.flush()
                recved_size += len(data)
            file.close()

        if not request["block"]:
            self._put_result(session_id)

        self._make_signal(session_id)

    thread = Thread(target=session)
    thread.start()
    signal = self._recv_signal(session_id)
    if signal["cancel"] and thread.is_alive():
        thread.kill()
        thread.join()
        if Config.debug:
            self._respond_ok(session_id, cancel=True, last_one=True,
                             debug="Cancel.")
        else:
            self._respond_ok(session_id, cancel=True, last_one=True)

        self._put_result(session_id, cancelled=True)
    else:
        thread.join()


def init_file_transfer_methods(cls):
    cls.get_file = get_file
    cls._process_get_file = _process_get_file
    cls.put_file = put_file
    cls._process_put_file = _process_put_file
    cls.get_folder = get_folder
    cls._process_get_folder = _process_get_folder
    cls.put_folder = put_folder
    cls._process_put_folder = _process_put_folder
    cls._write_to_file = _write_to_file
    cls._process__write_to_file = _process__write_to_file
    cls._close_file = _close_file
    cls._process__close_file = _process__close_file

    return cls



from .utils import eprint


def _server_get(self, timeout=None, block=True):
    session_id = self._get_session_id()
    self._request(session_id, timeout=timeout, block=block)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["value"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_get(self, session_id, request):
    try:
        value = self._parent.get(
            timeout=request["data"]["timeout"], block=request["block"]
        )
        self._respond_ok(session_id, value=value, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_put(self, value, timeout=None, block=True):
    session_id = self._get_session_id()
    self._request(
        session_id, value=value,
        timeout=timeout, block=block, last_one=True
    )
    response = self._recv_response(session_id)

    if not response["success"]:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_put(self, session_id, request):
    try:
        data = request["data"]
        self._parent.put(
            data["value"], timeout=data["timeout"], block=request["block"]
        )
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_qsize(self, timeout=None, block=True):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["qsize"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_qsize(self, session_id, request):
    try:
        qsize = self._parent.qsize()
        self._respond_ok(session_id, qsize=qsize, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queue_get(self, name, timeout=None, block=True):
    session_id = self._get_session_id()
    self._request(session_id, name=name, timeout=timeout, block=block)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["value"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queue_get(self, session_id, request):
    try:
        data = request["data"]
        value = self._parent.queues[data["name"]].get(
            timeout=data["timeout"], block=request["block"]
        )
        self._respond_ok(session_id, value=value, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queue_put(self, name, value, timeout=None, block=True):
    session_id = self._get_session_id()
    self._request(
        session_id, name=name, value=value,
        timeout=timeout, block=block
    )
    response = self._recv_response(session_id)

    if not response["success"]:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queue_put(self, session_id, request):
    try:
        data = request["data"]
        self._parent.queues[data["name"]].put(
            data["value"], timeout=data["timeout"], block=request["block"]
        )
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queue_len(self, name):
    session_id = self._get_session_id()
    self._request(session_id, name=name)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["len"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queue_len(self, session_id, request):
    try:
        length = 0
        name = request["data"]["name"]
        if name in self._parent.queues:
            length = self._parent.queues[name].qsize()

        self._respond_ok(session_id, len=length, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queues_delitem(self, name):
    session_id = self._get_session_id()
    self._request(session_id, name=name)
    response = self._recv_response(session_id)

    if not response["success"]:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queues_delitem(self, session_id, request):
    try:
        del self._parent.queues[request["data"]["name"]]
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queues_len(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["len"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queues_len(self, session_id, request):
    try:
        self._respond_ok(
            session_id, len=len(self._parent.queues), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queues_contains(self, name):
    session_id = self._get_session_id()
    self._request(session_id, name=name)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["contains"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queues_contains(self, session_id, request):
    try:
        contains = (request["data"]["name"] in self._parent.queues)
        self._respond_ok(
            session_id, contains=contains, last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queues_keys(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["keys"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queues_keys(self, session_id, request):
    try:
        self._respond_ok(
            session_id, keys=self._parent.queues.keys(), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queues_iter(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if response["success"]:
        return response["data"]["iter"]
    else:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queues_iter(self, session_id, request):
    try:
        self._respond_ok(
            session_id, iter=iter(self._parent.queues), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _server_queues_clear(self):
    session_id = self._get_session_id()
    self._request(session_id)
    response = self._recv_response(session_id)

    if not response["success"]:
        eprint(response["traceback"])
        raise response["exception"]


def _process__server_queues_clear(self, session_id, request):
    try:
        self._parent.queues.clear()
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def init_global_queues_methods(cls):
    cls._server_get = _server_get
    cls._server_put = _server_put
    cls._server_qsize = _server_qsize
    cls._server_queue_get = _server_queue_get
    cls._server_queue_put = _server_queue_put
    cls._server_queue_len = _server_queue_len
    cls._server_queues_len = _server_queues_len
    cls._server_queues_delitem = _server_queues_delitem
    cls._server_queues_contains = _server_queues_contains
    cls._server_queues_keys = _server_queues_keys
    cls._server_queues_iter = _server_queues_iter
    cls._server_queues_clear = _server_queues_clear
    cls._process__server_get = _process__server_get
    cls._process__server_put = _process__server_put
    cls._process__server_qsize = _process__server_qsize
    cls._process__server_queue_get = _process__server_queue_get
    cls._process__server_queue_put = _process__server_queue_put
    cls._process__server_queue_len = _process__server_queue_len
    cls._process__server_queues_len = _process__server_queues_len
    cls._process__server_queues_delitem = _process__server_queues_delitem
    cls._process__server_queues_contains = _process__server_queues_contains
    cls._process__server_queues_keys = _process__server_queues_keys
    cls._process__server_queues_iter = _process__server_queues_iter
    cls._process__server_queues_clear = _process__server_queues_clear

    return cls



from .CloseableQueue import CloseableQueue


class CloseablePipePeer:

    def __init__(self):
        self._other_peer = None
        self._queue = CloseableQueue()

    def send(self, var, block=True, timeout=None):
        self._other_peer._queue.put(var, block, timeout)

    def recv(self, block=True, timeout=None):
        return self._queue.get(block, timeout)

    def close(self):
        self._queue.close()
        self._other_peer._queue.close()

    def closed(self):
        return self._queue.closed()


def CloseablePipe():
    peer1 = CloseablePipePeer()
    peer2 = CloseablePipePeer()
    peer1._other_peer = peer2
    peer2._other_peer = peer1
    return peer1, peer2



import threading
import dill
import sys
import traceback
import uuid
import socket
from concurrent.futures import ThreadPoolExecutor

from .CloseableQueue import CloseableQueue
from .CloseablePipe import CloseablePipe
from .NodeInternalClasses import Pipes, Queues, PipeDict, QueueDict, ServerPeer

complex_request = ["get_file", "put_file",
                   "get_folder", "put_folder",
                   "eval", "exec", "system",
                   "call", "call_remote",
                   "execfile", "exec_remote_file"]


def __init__(self, connection, server=None):
    self._connection = connection
    if hasattr(self, "_send_buffer"):
        self._connection.setsockopt(
            socket.SOL_SOCKET, socket.SO_SNDBUF, self._send_buffer
        )
    else:
        self._send_buffer = self._connection.getsockopt(
            socket.SOL_SOCKET, socket.SO_SNDBUF
        )

    if hasattr(self, "_recv_buffer"):
        self._connection.setsockopt(
            socket.SOL_SOCKET, socket.SO_RCVBUF, self._recv_buffer
        )
    else:
        self._recv_buffer = self._connection.getsockopt(
            socket.SOL_SOCKET, socket.SO_RCVBUF
        )

    self._parent = server
    self._pipes = Pipes(self)
    self._queues = Queues(self)
    self._result_queues_for_future = Queues(self, private=True)
    self._recved_binary_queue = CloseableQueue()

    if self._parent is None:
        self._local_vars = {}
        self._pipe = CloseablePipe()
        self._queue = CloseableQueue()
        self._actual_pipes = PipeDict()
        self._actual_queues = QueueDict()
        self._actual_result_queues_for_future = QueueDict()
        self._server_peer = ServerPeer(self)
    else:
        self._server_peer = self._parent

    self._recving_thread = None
    self._decoding_thread = None

    self._recved_responses = QueueDict()
    self._recved_signals = QueueDict()

    self._out_file = None

    self._simple_threads_pool = None
    self._complex_threads_pool = None


def _start(self):
    if self._simple_threads_pool is None:
        self._simple_threads_pool = ThreadPoolExecutor(max_workers=5)

    if self._complex_threads_pool is None:
        self._complex_threads_pool = ThreadPoolExecutor(max_workers=1)

    if self._recving_thread is None or \
       not self._recving_thread.is_alive():
        self._recving_thread = threading.Thread(
            target=self._recving_loop, daemon=True
        )
        self._recving_thread.start()

    if self._decoding_thread is None or \
       not self._decoding_thread.is_alive():
        self._decoding_thread = threading.Thread(
            target=self._decoding_loop, daemon=True
        )
        self._decoding_thread.start()


def __del__(self):
    try:
        self.close()
    except BaseException:
        pass


def hold_on(self):
    threading.Event().wait()


def close(self):
    if self._parent is None:
        try:
            self._request(session_id=self._get_session_id())
        except BaseException:
            pass

        try:
            self._connection.close()
        except BaseException:
            pass

        self._connection = None

        try:
            self._queue.close()
        except BaseException:
            pass

        try:
            self._recved_binary_queue.close()
        except BaseException:
            pass

        try:
            self._pipe[0].close()
        except BaseException:
            pass

        try:
            self._pipe[1].close()
        except BaseException:
            pass

        try:
            self._actual_queues.clear()
        except BaseException:
            pass

        try:
            self._actual_result_queues_for_future.clear()
        except BaseException:
            pass

        try:
            self._actual_pipes.clear()
        except BaseException:
            pass

    else:
        try:
            self._connection.close()
        except BaseException:
            pass

        self._connection = None

        try:
            if self.address is not None:
                result = self._parent._callback_threads_pool.submit(
                    self._parent._on_disconnected, self.address)
                self._parent.disconnect_result[self.address] = result
        except BaseException:
            pass

        try:
            self._parent._connected_nodes.pop(self.address)
        except BaseException:
            pass

    if self._simple_threads_pool is not None:
        try:
            self._simple_threads_pool.shutdown()
        except BaseException:
            pass
        self._simple_threads_pool = None

    if self._complex_threads_pool is not None:
        try:
            self._complex_threads_pool.shutdown()
        except BaseException:
            pass
        self._complex_threads_pool = None

    try:
        self._recved_signals.clear()
    except BaseException:
        pass

    try:
        self._recved_responses.clear()
    except BaseException:
        pass


def _get_session_id(self):
    return uuid.uuid4().bytes


def _close_session(self, session_id):
    if session_id in self._recved_signals:
        del self._recved_signals[session_id]

    if session_id in self._recved_responses:
        del self._recved_responses[session_id]

    if self._parent is None:
        del self._actual_result_queues_for_future[session_id]


def _traceback(self, remote=True):
    exception_list = traceback.format_stack()
    exception_list = exception_list[:-2]
    exception_list.extend(traceback.format_tb(sys.exc_info()[2]))
    exception_list.extend(
        traceback.format_exception_only(
            sys.exc_info()[0], sys.exc_info()[1]
        )
    )

    exception_str = "Traceback (most recent call last):\n"
    if remote:
        exception_str = f"Remote {self.server_address} {exception_str}"

    exception_str += "".join(exception_list)

    return exception_str


def _process_close(self, request):
    self.close()


def _send(self, session_id, type, value):
    binary = dill.dumps(value, 3)
    type_bytes = type.to_bytes(1, byteorder='little', signed=False)
    length_bytes = len(binary).to_bytes(4, byteorder='little', signed=False)
    self._connection.sendall(session_id + type_bytes + length_bytes + binary)


def _send_signal(self, session_id, **kwargs):
    cancel = False
    if "cancel" in kwargs:
        cancel = kwargs.pop("cancel")

    signal = {
        "cancel": cancel,
        "data": kwargs
    }
    self._send(session_id, 2, signal)  # signal


def _request(self, session_id, **kwargs):
    block = True
    if "block" in kwargs:
        block = kwargs.pop("block")

    if "debug" in kwargs:
        print("request:", kwargs["debug"], flush=True)

    func = sys._getframe().f_back.f_code.co_name
    if "func" in kwargs:
        func = kwargs.pop("func")

    request = {
        "func": func,
        "block": block,
        "data": kwargs
    }
    self._send(session_id, 0, request)


def _respond_ok(self, session_id, **kwargs):
    block = True
    if "block" in kwargs:
        block = kwargs.pop("block")

    cancel = False
    if "cancel" in kwargs:
        cancel = kwargs.pop("cancel")

    last_one = False
    if "last_one" in kwargs:
        last_one = kwargs.pop("last_one")

    if "debug" in kwargs:
        print("respond:", kwargs["debug"], flush=True)

    if block:
        response = {
            "success": (not cancel),
            "cancel": cancel,
            "last_one": last_one,
            "data": kwargs
        }
        self._send(session_id, 1, response)
        if last_one:
            self._close_session(session_id)
    else:
        self._put_result(session_id, **kwargs)


def _respond_exception(self, session_id, exception, **kwargs):
    block = True
    if "block" in kwargs:
        block = kwargs.pop("block")

    if "debug" in kwargs:
        print("respond exception:", kwargs["debug"], flush=True)

    if block:
        response = {
            "success": False,
            "exception": exception,
            "traceback": self._traceback(),
            "last_one": True,
            "data": kwargs
        }
        self._send(session_id, 1, response)
        self._close_session(session_id)
    else:
        self._put_exception(session_id, exception, **kwargs)


def _put_result(self, session_id, **kwargs):
    result = {
        "success": True,
        "return_value": None,
        "exception": None,
        "traceback": None,
        "cancelled": False
    }
    for key in result:
        if key in kwargs:
            result[key] = kwargs.pop(key)
    result["data"] = kwargs

    self._result_queues_for_future[session_id].put(result)


def _put_exception(self, session_id, exception, **kwargs):
    result = {
        "success": False,
        "return_value": None,
        "exception": exception,
        "traceback": self._traceback(),
        "cancelled": False
    }
    for key in result:
        if key in kwargs:
            result[key] = kwargs.pop(key)
    result["data"] = kwargs

    self._result_queues_for_future[session_id].put(result)


def _make_signal(self, session_id, **kwargs):
    cancel = False
    if "cancel" in kwargs:
        cancel = kwargs.pop("cancel")

    signal = {
        "cancel": cancel,
        "data": kwargs
    }
    self._recved_signals[session_id].put(signal)


def _recv_signal(self, session_id):
    signal = self._recved_signals[session_id].get()

    if "debug" in signal["data"]:
        print("recved signal:", signal["data"]["debug"])

    return signal


def _recv_response(self, session_id):
    response = self._recved_responses[session_id].get()

    if "debug" in response["data"]:
        print("recved response:", response["data"]["debug"])

    if response["last_one"]:
        self._close_session(session_id)

    return response


def _recving_loop(self):
    while True:
        try:
            recved_binary = self._connection.recv(self.recv_buffer)
            self._recved_binary_queue.put(recved_binary)
        except BaseException:
            self.close()
            return


def _decoding_loop(self):
    buffer = b''
    while True:
        while len(buffer) < 21:
            try:
                buffer += self._recved_binary_queue.get()
            except BaseException:
                return
        session_id = buffer[:16]
        message_type = buffer[16]
        var_length = int.from_bytes(
            buffer[17:21], byteorder="little", signed=False
        )
        stop = 21 + var_length
        while len(buffer) < stop:
            try:
                buffer += self._recved_binary_queue.get()
            except BaseException:
                return

        var_bytes = buffer[21:stop]
        buffer = buffer[stop:]
        var = dill.loads(var_bytes)
        if message_type == 0:  # request
            try:
                print(var["data"]["debug"])
            except BaseException:
                pass

            var_func = var["func"]
            process_func = getattr(self, "_process_" + var_func)
            if var_func in complex_request:
                self._complex_threads_pool.submit(
                    process_func, session_id, var
                )
            else:
                if var_func in ["_write_to_file", "_close_file"]:
                    process_func(session_id, var)
                else:
                    self._simple_threads_pool.submit(
                        process_func, session_id, var
                    )
        elif message_type == 1:  # response
            self._recved_responses[session_id].put(var)
        elif message_type == 2:  # signal
            self._recved_signals[session_id].put(var)


def init_basic_methods(cls):
    cls.__init__ = __init__
    cls.__del__ = __del__
    cls.close = close
    cls.hold_on = hold_on
    cls._start = _start
    cls._get_session_id = _get_session_id
    cls._traceback = _traceback
    cls._process_close = _process_close
    cls._send = _send
    cls._send_signal = _send_signal
    cls._request = _request
    cls._respond_ok = _respond_ok
    cls._respond_exception = _respond_exception
    cls._put_result = _put_result
    cls._put_exception = _put_exception
    cls._make_signal = _make_signal
    cls._recv_signal = _recv_signal
    cls._recv_response = _recv_response
    cls._recving_loop = _recving_loop
    cls._decoding_loop = _decoding_loop
    cls._close_session = _close_session
    return cls



from .utils import eprint


def send(self, value):
    if self._parent is None:
        self._pipe[1].send(value)
    else:
        session_id = self._get_session_id()
        self._request(session_id, value=value)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process_send(self, session_id, request):
    try:
        self._pipe[0].send(request["data"]["value"])
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def recv(self):
    if self._parent is None:
        return self._pipe[1].recv()
    else:
        session_id = self._get_session_id()
        self._request(session_id)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process_recv(self, session_id, request):
    try:
        self._respond_ok(session_id, value=self._pipe[0].recv(), last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_pipe_send(self, name, value):
    if self._parent is None:
        self._actual_pipes[name][1].send(value)
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name, value=value)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_pipe_send(self, session_id, request):
    try:
        data = request["data"]
        self._actual_pipes[data["name"]][0].send(data["value"])
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_pipe_recv(self, name):
    if self._parent is None:
        return self._actual_pipes[name][1].recv()
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_pipe_recv(self, session_id, request):
    try:
        name = request["data"]["name"]
        self._respond_ok(
            session_id, value=self._actual_pipes[name][0].recv(),
            last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_pipes_delitem(self, name):
    if self._parent is None:
        del self._actual_pipes[name]
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_pipes_delitem(self, session_id, request):
    try:
        del self._actual_pipes[request["data"]["name"]]
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_pipes_clear(self):
    if self._parent is None:
        self._actual_pipes.clear()
    else:
        session_id = self._get_session_id()
        self._request(session_id)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_pipes_clear(self, session_id, request):
    try:
        self._actual_pipes.clear()
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_pipes_len(self):
    if self._parent is None:
        return len(self._actual_pipes)
    else:
        session_id = self._get_session_id()
        self._request(session_id)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["len"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_pipes_len(self, session_id, request):
    try:
        self._respond_ok(
            session_id, len=len(self._actual_pipes), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_pipes_contains(self, name):
    if self._parent is None:
        return (name in self._actual_pipes)
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["contains"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_pipes_contains(self, session_id, request):
    try:
        pipes_contains = (request["data"]["name"] in self._actual_pipes)
        self._respond_ok(session_id, contains=pipes_contains, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_pipes_keys(self):
    if self._parent is None:
        return self._actual_pipes.keys()
    else:
        session_id = self._get_session_id()
        self._request(session_id)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["keys"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_pipes_keys(self, session_id, request):
    try:
        self._respond_ok(
            session_id, keys=self._actual_pipes.keys(), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_pipes_iter(self):
    if self._parent is None:
        return self._actual_pipes.__iter__()
    else:
        session_id = self._get_session_id()
        self._request(session_id)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["iter"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_pipes_iter(self, session_id, request):
    try:
        self._respond_ok(
            session_id, iter=self._actual_pipes.__iter__(), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def init_local_pipes_methods(cls):
    cls.send = send
    cls.recv = recv
    cls._locals_pipe_send = _locals_pipe_send
    cls._locals_pipe_recv = _locals_pipe_recv
    cls._locals_pipes_len = _locals_pipes_len
    cls._locals_pipes_contains = _locals_pipes_contains
    cls._locals_pipes_keys = _locals_pipes_keys
    cls._locals_pipes_iter = _locals_pipes_iter
    cls._locals_pipes_delitem = _locals_pipes_delitem
    cls._locals_pipes_clear = _locals_pipes_clear
    cls._process_send = _process_send
    cls._process_recv = _process_recv
    cls._process__locals_pipe_send = _process__locals_pipe_send
    cls._process__locals_pipe_recv = _process__locals_pipe_recv
    cls._process__locals_pipes_len = _process__locals_pipes_len
    cls._process__locals_pipes_contains = _process__locals_pipes_contains
    cls._process__locals_pipes_keys = _process__locals_pipes_keys
    cls._process__locals_pipes_iter = _process__locals_pipes_iter
    cls._process__locals_pipes_delitem = _process__locals_pipes_delitem
    cls._process__locals_pipes_clear = _process__locals_pipes_clear

    return cls



from .utils import eprint


def put(self, value, timeout=None, block=True):
    if self._parent is None:
        self._queue.put(value)
    else:
        session_id = self._get_session_id()
        self._request(session_id, value=value, timeout=timeout, block=block)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process_put(self, session_id, request):
    try:
        data = request["data"]
        self._queue.put(
            data["value"], timeout=data["timeout"], block=request["block"]
        )
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def get(self, timeout=None, block=True):
    if self._parent is None:
        return self._queue.get()
    else:
        session_id = self._get_session_id()
        self._request(session_id, timeout=timeout, block=block)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process_get(self, session_id, request):
    try:
        value = self._queue.get(
            timeout=request["data"]["timeout"], block=request["block"]
        )
        self._respond_ok(session_id, value=value, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def qsize(self):
    if self._parent is None:
        return self._queue.qsize()
    else:
        session_id = self._get_session_id()
        self._request(session_id)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["qsize"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process_qsize(self, session_id, request):
    try:
        qsize = self._queue.qsize()
        self._respond_ok(session_id, qsize=qsize, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _used_queues(self, is_private):
    if not is_private:
        return self._actual_queues
    else:
        return self._actual_result_queues_for_future


def _locals_queue_put(self, name, value, timeout, block, is_private):
    if self._parent is None:
        used_queue = self._used_queues(is_private)[name]
        used_queue.put(value, timeout=timeout, block=block)
    else:
        session_id = self._get_session_id()
        self._request(
            session_id, name=name, value=value, timeout=timeout,
            block=block, is_private=is_private
        )
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queue_put(self, session_id, request):
    try:
        data = request["data"]
        used_queue = self._used_queues(data["is_private"])[data["name"]]
        used_queue.put(
            data["value"], timeout=data["timeout"], block=request["block"]
        )
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_queue_get(self, name, timeout, block, is_private):
    if self._parent is None:
        used_queue = self._used_queues(is_private)[name]
        return used_queue.get(timeout=timeout, block=block)
    else:
        session_id = self._get_session_id()
        self._request(
            session_id, name=name, timeout=timeout,
            block=block, is_private=is_private
        )
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queue_get(self, session_id, request):
    try:
        data = request["data"]
        used_queue = self._used_queues(data["is_private"])[data["name"]]
        value = used_queue.get(timeout=data["timeout"], block=request["block"])
        self._respond_ok(session_id, value=value, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_queue_len(self, name, is_private):
    if self._parent is None:
        used_queues = self._used_queues(is_private)
        if name not in used_queues:
            return 0
        else:
            return used_queues[name].qsize()
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name, is_private=is_private)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["len"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queue_len(self, session_id, request):
    try:
        data = request["data"]
        name = data["name"]
        used_queues = self._used_queues(data["is_private"])
        queue_len = 0
        if name in used_queues:
            queue_len = used_queues[name].qsize()
        self._respond_ok(session_id, len=queue_len, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_queues_delitem(self, name, is_private):
    if self._parent is None:
        del self._used_queues(is_private)[name]
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name, is_private=is_private)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queues_delitem(self, session_id, request):
    try:
        data = request["data"]
        del self._used_queues(data["is_private"])[data["name"]]
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_queues_len(self, is_private):
    if self._parent is None:
        return len(self._used_queues(is_private))
    else:
        session_id = self._get_session_id()
        self._request(session_id, is_private=is_private)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["len"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queues_len(self, session_id, request):
    try:
        queues_len = len(self._used_queues(request["data"]["is_private"]))
        self._respond_ok(session_id, len=queues_len, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_queues_contains(self, name, is_private):
    if self._parent is None:
        return (name in self._used_queues(is_private))
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name, is_private=is_private)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["contains"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queues_contains(self, session_id, request):
    try:
        data = request["data"]
        used_queues = self._used_queues(data["is_private"])
        queues_conntains = (data["name"] in used_queues)
        self._respond_ok(session_id, contains=queues_conntains, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_queues_keys(self, is_private):
    if self._parent is None:
        return self._used_queues(is_private).keys()
    else:
        session_id = self._get_session_id()
        self._request(session_id, is_private=is_private)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["keys"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queues_keys(self, session_id, request):
    try:
        queues_keys = self._used_queues(request["data"]["is_private"]).keys()
        self._respond_ok(session_id, keys=queues_keys, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_queues_iter(self, is_private):
    if self._parent is None:
        return self._used_queues(is_private).__iter__()
    else:
        session_id = self._get_session_id()
        self._request(session_id, is_private=is_private)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["iter"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queues_iter(self, session_id, request):
    try:
        used_queues = self._used_queues(request["data"]["is_private"])
        queues_iter = used_queues.__iter__()
        self._respond_ok(session_id, iter=queues_iter, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def _locals_queues_clear(self, is_private):
    if self._parent is None:
        self._used_queues(is_private).clear()
    else:
        session_id = self._get_session_id()
        self._request(session_id, is_private=is_private)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process__locals_queues_clear(self, session_id, request):
    try:
        self._used_queues(request["data"]["is_private"]).clear()
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def init_local_queues_methods(cls):
    cls.put = put
    cls.get = get
    cls.qsize = qsize
    cls._used_queues = _used_queues
    cls._locals_queue_put = _locals_queue_put
    cls._locals_queue_get = _locals_queue_get
    cls._locals_queue_len = _locals_queue_len
    cls._locals_queues_len = _locals_queues_len
    cls._locals_queues_contains = _locals_queues_contains
    cls._locals_queues_keys = _locals_queues_keys
    cls._locals_queues_iter = _locals_queues_iter
    cls._locals_queues_delitem = _locals_queues_delitem
    cls._locals_queues_clear = _locals_queues_clear
    cls._process_put = _process_put
    cls._process_get = _process_get
    cls._process_qsize = _process_qsize
    cls._process__locals_queue_put = _process__locals_queue_put
    cls._process__locals_queue_get = _process__locals_queue_get
    cls._process__locals_queue_len = _process__locals_queue_len
    cls._process__locals_queues_len = _process__locals_queues_len
    cls._process__locals_queues_contains = _process__locals_queues_contains
    cls._process__locals_queues_keys = _process__locals_queues_keys
    cls._process__locals_queues_iter = _process__locals_queues_iter
    cls._process__locals_queues_delitem = _process__locals_queues_delitem
    cls._process__locals_queues_clear = _process__locals_queues_clear

    return cls



import socket

from .Node import Node
from .utils import get_ip


class Client(Node):

    def __init__(self, ip=None, port=None):
        if ip is None:
            ip = get_ip()
        if port is None:
            port = 0

        self._ip = ip
        self._port = port

        self._connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._connection.bind((ip, port))
        Node.__init__(self, self._connection)

    def connect(self, ip, port=None):
        if port is None:
            port = ip[1]
            ip = ip[0]

        server_address = self.server_address
        try:
            if server_address[0] == ip and \
               server_address[1] == port and \
               not self.is_closed:
                return
        except BaseException:
            pass

        if server_address is not None and not self.is_closed:
            self.close()

        if self._connection is None:
            self._connection = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM
            )
            self._connection.bind((self._ip, self._port))

        self._connection.connect((ip, port))
        Node.__init__(self, self._connection)
        self._start()
        self._recv_response(session_id=b'\x00'*16)



import socket
import threading
import os
from concurrent.futures import ThreadPoolExecutor

from .Node import Node
from .NodeInternalClasses import OrderedDict
from .utils import file_size, get_ip


class Peer(Node):
    def __init__(self, ip=None, port=None):
        self.connect_results = OrderedDict()
        self.disconnect_results = OrderedDict()
        self._callback_threads_pool = ThreadPoolExecutor(max_workers=5)

        self._on_connected = lambda client: None
        self._on_disconnect = lambda address: None

        self._connected_nodes = OrderedDict()

        self._continue_accept = False
        self._accept_thread = None

        if ip is None:
            ip = get_ip()
        if port is None:
            port = 0

        self._ip = ip
        self._port = port

        self._connection1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._connection1.bind((ip, 0))
        Node.__init__(self, self._connection1)

        self.start()

    @property
    def address(self):
        try:
            return self._connection2.getsockname()
        except BaseException:
            return None

    def set_connected_callback(self, func):
        self._on_connected = func

    def set_disconnected_callback(self, func):
        self._on_disconnected = func

    def connect(self, ip, port=None):
        server_address = ip
        if port is not None:
            server_address = (ip, port)

        if self._connection1 is None:
            self._connection1 = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM
            )
            self._connection1.bind((self._ip, 0))

        self._connection1.connect(server_address)
        Node._start(self)

    def start(self, ip=None, port=None):
        if self._continue_accept:
            return

        if ip is None:
            ip = self._ip
        if port is None:
            port = self._port

        if ip is None:
            ip = get_ip()
        if port is None:
            port = 0

        self._ip = ip
        self._port = port

        self._connection2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._connection2.bind((self._ip, self._port))
        self._connection2.listen(5)

        self._continue_accept = True
        if self._accept_thread is None or not self._accept_thread.is_alive():
            self._accept_thread = threading.Thread(
                target=self._accepting_loop, daemon=True
            )
            self._accept_thread.start()

    def close(self):
        self._continue_accept = False

        try:
            keys = self._connected_nodes.keys()
            for address in keys:
                try:
                    self._connected_nodes[address].close()
                except BaseException:
                    pass
        except BaseException:
            pass

        self._connected_nodes = OrderedDict()

        self.connect_results = OrderedDict()
        self.disconnect_results = OrderedDict()
        self._callback_threads_pool.shutdown()
        self._callback_threads_pool = ThreadPoolExecutor(max_workers=5)

        if self._accept_thread is not None and self._accept_thread.is_alive():
            self._accept_thread.join()
        self._accept_thread = None

        try:
            Node.close(self)
        except BaseException:
            pass

        try:
            self._connection2.close()
        except BaseException:
            pass

    @property
    def clients(self):
        return self._connected_nodes

    def __del__(self):
        try:
            self.close()
        except BaseException:
            pass

    def put_file_to(self, src_file_name, address_file_map):
        if not os.path.isfile(src_file_name):
            raise FileNotFoundError(f"File {src_file_name} is not exists.")

        sent_size = 0
        src_file_size = file_size(src_file_name)
        with open(src_file_name, "rb") as file:
            while sent_size < src_file_size:
                data = file.read(self.send_buffer)
                for address in address_file_map:
                    try:
                        client = self._connected_nodes[address]
                        client._write_to_file(data, address_file_map[address])
                    except BaseException:
                        pass
                sent_size += len(data)

        for address in address_file_map:
            try:
                client = self._connected_nodes[address]
                client.close_file()
            except BaseException:
                pass

    def put_folder_to(self, src_folder_name, address_file_map):
        if not os.path.isdir(src_folder_name):
            raise FileNotFoundError(f"Folder {src_folder_name} is not exists.")

        i = len(src_folder_name)-1
        while src_folder_name[i] in ['/', '\\']:
            i -= 1

        for root, dirs, files in os.walk(src_folder_name):
            for name in files:
                src_file_name = os.path.join(root, name)

                sent_size = 0
                src_file_size = file_size(src_file_name)
                with open(src_file_name, "rb") as file:
                    while sent_size < src_file_size:
                        data = file.read(self.send_buffer)
                        for address in address_file_map:
                            try:
                                client = self._connected_nodes[address]
                                client._write_to_file(
                                    data, address_file_map[address]
                                )
                            except BaseException:
                                pass
                        sent_size += len(data)

        for address in address_file_map:
            try:
                client = self._connected_nodes[address]
                client._close_file()
            except BaseException:
                pass

    def hold_on(self):
        threading.Event().wait()

    def _accepting_loop(self):
        while self._continue_accept:
            try:
                client_socket, address = self._connection2.accept()
                node = Node(client_socket, self)
                node._start()
                self._connected_nodes[address] = node

                result = self._callback_threads_pool.submit(
                    self._on_connected, self._connected_nodes[address])
                self.connect_results[address] = result
            except BaseException:
                break



from .NodeBasicMethods import init_basic_methods
from .NodeLocalPipesMethods import init_local_pipes_methods
from .NodeLocalQueuesMethods import init_local_queues_methods
from .NodeLocalDictMethods import init_local_dict_methods
from .NodeGlobalQueuesMethods import init_global_queues_methods
from .NodeGlobalDictMethods import init_global_dict_methods
from .NodeFileTransferMethods import init_file_transfer_methods
from .NodeRPCMethods import init_rpc_methods

import socket


@init_rpc_methods
@init_file_transfer_methods
@init_global_dict_methods
@init_global_queues_methods
@init_local_dict_methods
@init_local_queues_methods
@init_local_pipes_methods
@init_basic_methods
class Node:

    @property
    def is_closed(self):
        try:
            return self._connection._closed
        except BaseException:
            return True

    @property
    def address(self):
        try:
            if self._parent is None:
                return self._connection.getsockname()
            else:
                return self._connection.getpeername()
        except BaseException:
            return None

    @property
    def server_address(self):
        try:
            if self._parent is None:
                return self._connection.getpeername()
            else:
                return self._parent.address
        except BaseException:
            return None

    @property
    def server(self):
        return self._server_peer

    @property
    def queues(self):
        return self._queues

    @property
    def pipes(self):
        return self._pipes

    @property
    def send_buffer(self):
        return self._send_buffer

    @send_buffer.setter
    def send_buffer(self, buffer_size):
        self._connection.setsockopt(
            socket.SOL_SOCKET, socket.SO_SNDBUF, buffer_size
        )
        self._send_buffer = self._connection.getsockopt(
            socket.SOL_SOCKET, socket.SO_SNDBUF
        )
        if self._parent is not None:
            self._parent._send_buffer = self._send_buffer

    @property
    def recv_buffer(self):
        return self._recv_buffer

    @recv_buffer.setter
    def recv_buffer(self, buffer_size):
        self._connection.setsockopt(
            socket.SOL_SOCKET, socket.SO_RCVBUF, buffer_size
        )
        self._recv_buffer = self._connection.getsockopt(
            socket.SOL_SOCKET, socket.SO_RCVBUF
        )
        if self._parent is not None:
            self._parent._recv_buffer = self._recv_buffer



import os
import subprocess

from .NodeInternalClasses import Future, Thread, AutoSendBuffer
from .utils import eprint


def eval(self, statement, block=True):
    session_id = self._get_session_id()

    self._request(session_id, statement=statement, block=block)
    if block:
        response = self._recv_response(session_id)

        if response["success"]:
            return response["data"]["return_value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]
    else:
        return Future(self, session_id)


def _process_eval(self, session_id, request):
    try:
        def session():
            block = request["block"]
            try:
                return_value = eval(request["data"]["statement"])
                self._respond_ok(
                    session_id, return_value=return_value,
                    block=block, last_one=True
                )
            except BaseException as e:
                self._respond_exception(session_id, e, block=block)
            self._make_signal(session_id)

        thread = Thread(target=session)
        thread.start()
        signal = self._recv_signal(session_id)
        if signal["cancel"] and thread.is_alive():
            thread.kill()
            thread.join()
            self._put_result(session_id, cancelled=True)
        else:
            thread.join()

    except BaseException as e:
        self._respond_exception(session_id, e, block=request["block"])


def exec(self, code, block=True):
    session_id = self._get_session_id()

    self._request(session_id, code=code, block=block)
    if block:
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]
    else:
        return Future(self, session_id)


def _process_exec(self, session_id, request):
    try:
        def session():
            try:
                exec(request["data"]["code"], globals())
                self._respond_ok(
                    session_id, block=request["block"], last_one=True
                )
            except BaseException as e:
                self._respond_exception(session_id, e, block=request["block"])

            self._make_signal(session_id)

        thread = Thread(target=session)
        thread.start()
        signal = self._recv_signal(session_id)
        if signal["cancel"] and thread.is_alive():
            thread.kill()
            thread.join()
            self._put_result(session_id, cancelled=True)
        else:
            thread.join()
    except BaseException as e:
        self._respond_exception(session_id, e, block=request["block"])


def system(self, cmd,
           quiet=False, remote_quiet=False,
           once_all=False, block=True):
    session_id = self._get_session_id()

    self._request(
        session_id, cmd=cmd, remote_quiet=remote_quiet,
        once_all=once_all, block=block
    )
    if not once_all:
        def session(session_id, quiet):
            while True:
                response = self._recv_response(session_id)
                if response["success"]:
                    if not response["data"]["end"]:
                        if not quiet:
                            print(response["data"]["stdout"],
                                  end="", flush=True)
                    else:
                        return response["data"]["return_value"]
                else:
                    eprint(response["traceback"])
                    raise response["exception"]

        if block:
            session(session_id, quiet)
        else:
            thread = Thread(target=session, args=(session_id, quiet))
            thread.start()
            return Future(self, session_id)
    else:
        if block:
            response = self._recv_response(session_id)
            if response["success"]:
                if not quiet:
                    print(response["data"]["stdout"], end="", flush=True)
                    return response["data"]["return_value"]
            else:
                eprint(response["traceback"])
                raise response["exception"]
        else:
            return Future(self, session_id)


def _process_system(self, session_id, request):
    try:
        if not request["data"]["once_all"]:
            process = subprocess.Popen(
                request["data"]["cmd"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True
            )
            self.__sent_end = False
            self.__did_put_result = False

            def session():
                remote_quiet = request["data"]["remote_quiet"]
                auto_send_buffer = AutoSendBuffer(
                    self, session_id, timeout=0.1, quiet=remote_quiet
                )
                while True:
                    byte = process.stdout.read(1)
                    if byte == b'':
                        break
                    auto_send_buffer.append(byte)
                auto_send_buffer.stop()
                process.communicate()
                self._respond_ok(
                    session_id, end=True,
                    return_value=process.returncode, last_one=True
                )
                self.__sent_end = True
                if not request["block"]:
                    self._put_result(
                        session_id, return_value=process.returncode
                    )
                    self.__did_put_result = True
                self._make_signal(session_id)

            thread = Thread(target=session)
            thread.start()
            signal = self._recv_signal(session_id)
            if signal["cancel"] and thread.is_alive():
                process.kill()
                thread.kill()
                thread.join()
                if not self.__sent_end:
                    self._respond_ok(
                        session_id, end=True, return_value=0,
                        block=request["block"], last_one=True
                    )
                if not self.__did_put_result and not request["block"]:
                    self._put_result(session_id, cancel=True)
            else:
                thread.join()
        else:
            process = subprocess.Popen(
                request["data"]["cmd"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True
            )

            def session():
                stdout, stderr = process.communicate()

                stdout = stdout.decode("utf-8")
                if not request["data"]["remote_quiet"]:
                    print(stdout, end="", flush=True)

                self._respond_ok(
                    session_id, stdout=stdout, return_value=process.returncode,
                    block=request["block"], last_one=True
                )
                self._make_signal(session_id)

            thread = Thread(target=session)
            thread.start()
            signal = self._recv_signal(session_id)
            if signal["cancel"] and thread.is_alive():
                process.kill()
                thread.kill()
                thread.join()
                self._put_result(session_id, cancelled=True)
            else:
                thread.join()

    except BaseException as e:
        self._respond_exception(
            session_id, e,
            block=(request["block"] or not request["data"]["once_all"])
        )


# call local function on remote computer
def call(self, target, args=(), kwargs={}, block=True):
    session_id = self._get_session_id()

    self._request(
        session_id, target=target,
        args=args, kwargs=kwargs, block=block
    )
    if block:
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["return_value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]
    else:
        return Future(self, session_id)


def _process_call(self, session_id, request):
    try:
        def session():
            try:
                data = request["data"]
                target = data["target"]
                args = data["args"]
                kwargs = data["kwargs"]
                return_value = target(*args, **kwargs)
                self._respond_ok(
                    session_id, return_value=return_value,
                    block=request["block"], last_one=True
                )
            except BaseException as e:
                self._respond_exception(session_id, e, block=request["block"])

            self._make_signal(session_id)

        thread = Thread(target=session)
        thread.start()
        signal = self._recv_signal(session_id)
        if signal["cancel"] and thread.is_alive():
            thread.kill()
            thread.join()
            self._put_result(session_id, cancelled=True)
        else:
            thread.join()
    except BaseException as e:
        self._respond_exception(session_id, e, block=request["block"])


# call remote function on remote computer
def call_remote(self, target, args=(), kwargs={}, block=True):
    session_id = self._get_session_id()

    self._request(
        session_id, target=target,
        args=args, kwargs=kwargs, block=block
    )
    if block:
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["return_value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]
    else:
        return Future(self, session_id)


def _process_call_remote(self, session_id, request):
    try:
        def session():
            try:
                data = request["data"]
                target = eval(data["target"])
                args = data["args"]
                kwargs = data["kwargs"]
                return_value = target(*args, **kwargs)
                self._respond_ok(
                    session_id, return_value=return_value,
                    block=request["block"], last_one=True
                )
            except BaseException as e:
                self._respond_exception(session_id, e, block=request["block"])

            self._make_signal(session_id)

        thread = Thread(target=session)
        thread.start()
        signal = self._recv_signal(session_id)
        if signal["cancel"] and thread.is_alive():
            thread.kill()
            thread.join()
            self._put_result(session_id, cancelled=True)
        else:
            thread.join()

    except BaseException as e:
        self._respond_exception(session_id, e, block=request["block"])


def execfile(self, script_name, block=True):
    temp_file_name = f".temp_scripts/{os.path.basename(script_name)}"
    self.put_file(script_name, temp_file_name)
    return self.exec_remote_file(temp_file_name, block=block)


def exec_remote_file(self, script_name, block=True):
    session_id = self._get_session_id()

    self._request(session_id, script_name=script_name, block=block)
    if block:
        response = self._recv_response(session_id)

        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]
    else:
        return Future(self, session_id)


def _process_exec_remote_file(self, session_id, request):
    try:
        def session():
            try:
                exec(open(request["data"]["script_name"]).read(), globals())
                self._respond_ok(
                    session_id, block=request["block"], last_one=True
                )
            except BaseException as e:
                self._respond_exception(session_id, e, block=request["block"])

            self._make_signal(session_id)

        thread = Thread(target=session)
        thread.start()
        signal = self._recv_signal(session_id)
        if signal["cancel"] and thread.is_alive():
            thread.kill()
            thread.join()
            self._put_result(session_id, cancelled=True)
        else:
            thread.join()

    except BaseException as e:
        self._respond_exception(session_id, e, block=request["block"])


def init_rpc_methods(cls):
    cls.eval = eval
    cls.exec = exec
    cls.execfile = execfile
    cls.exec_remote_file = exec_remote_file
    cls.system = system
    cls.call = call
    cls.call_remote = call_remote
    cls._process_system = _process_system
    cls._process_eval = _process_eval
    cls._process_exec = _process_exec
    cls._process_exec_remote_file = _process_exec_remote_file
    cls._process_call = _process_call
    cls._process_call_remote = _process_call_remote

    return cls



from .utils import eprint


def __getitem__(self, name):
    if self._parent is None:
        return self._local_vars[name]
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process___getitem__(self, session_id, request):
    try:
        value = self._local_vars[request["data"]["name"]]
        self._respond_ok(session_id, value=value, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def __setitem__(self, name, value):
    if self._parent is None:
        self._local_vars[name] = value
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name, value=value)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process___setitem__(self, session_id, request):
    try:
        data = request["data"]
        self._local_vars[data["name"]] = data["value"]
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def __delitem__(self, name):
    if self._parent is None:
        del self._local_vars[name]
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name)
        response = self._recv_response(session_id)
        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process___delitem__(self, session_id, request):
    try:
        del self._local_vars[request["data"]["name"]]
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def __iter__(self):
    if self._parent is None:
        return self._local_vars.__iter__()
    else:
        session_id = self._get_session_id()
        self._request(session_id)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["iter"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process___iter__(self, session_id, request):
    try:
        self._respond_ok(
            session_id, iter=iter(self._local_vars), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def __contains__(self, name):
    if self._parent is None:
        return self._local_vars.__contains__(name)
    else:
        session_id = self._get_session_id()
        self._request(session_id, name=name)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["contains"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process___contains__(self, session_id, request):
    try:
        contains = (request["data"]["name"] in self._local_vars)
        self._respond_ok(session_id, contains=contains, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def __len__(self):
    if self._parent is None:
        return self._local_vars.__len__()
    else:
        session_id = self._get_session_id()
        self._request(session_id)
        response = self._recv_response(session_id)
        if response["success"]:
            return response["data"]["len"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process___len__(self, session_id, request):
    try:
        self._respond_ok(
            session_id, len=len(self._local_vars), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def clear(self):
    if self._parent is None:
        self._local_vars.clear()
    else:
        session_id = self._get_session_id()

        self._request(session_id)
        response = self._recv_response(session_id)

        if not response["success"]:
            eprint(response["traceback"])
            raise response["exception"]


def _process_clear(self, session_id, request):
    try:
        self._local_vars.clear()
        self._respond_ok(session_id, last_one=True)
    except BaseException as e:
        self._respond_exception(session_id, e)


def keys(self):
    if self._parent is None:
        return self._local_vars.keys()
    else:
        session_id = self._get_session_id()

        self._request(session_id)
        response = self._recv_response(session_id)

        if response["success"]:
            return response["data"]["keys"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process_keys(self, session_id, request):
    try:
        self._respond_ok(
            session_id, keys=self._local_vars.keys(), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def values(self):
    if self._parent is None:
        return self._local_vars.values()
    else:
        session_id = self._get_session_id()

        self._request(session_id)
        response = self._recv_response(session_id)

        if response["success"]:
            return response["data"]["values"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process_values(self, session_id, request):
    try:
        self._respond_ok(
            session_id, values=self._local_vars.values(), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def items(self):
    if self._parent is None:
        return self._local_vars.items()
    else:
        session_id = self._get_session_id()

        self._request(session_id)
        response = self._recv_response(session_id)

        if response["success"]:
            return response["data"]["items"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process_items(self, session_id, request):
    try:
        self._respond_ok(
            session_id, items=self._local_vars.items(), last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def pop(self, name):
    if self._parent is None:
        return self._local_vars.pop(name)
    else:
        session_id = self._get_session_id()

        self._request(session_id, name=name)
        response = self._recv_response(session_id)

        if response["success"]:
            return response["data"]["value"]
        else:
            eprint(response["traceback"])
            raise response["exception"]


def _process_pop(self, session_id, request):
    try:
        self._respond_ok(
            session_id, value=self._local_vars.pop(request["data"]["name"]),
            last_one=True
        )
    except BaseException as e:
        self._respond_exception(session_id, e)


def init_local_dict_methods(cls):
    cls.__getitem__ = __getitem__
    cls.__setitem__ = __setitem__
    cls.__delitem__ = __delitem__
    cls.__iter__ = __iter__
    cls.__contains__ = __contains__
    cls.__len__ = __len__
    cls.clear = clear
    cls.keys = keys
    cls.values = values
    cls.items = items
    cls.pop = pop
    cls._process___getitem__ = _process___getitem__
    cls._process___setitem__ = _process___setitem__
    cls._process___delitem__ = _process___delitem__
    cls._process___iter__ = _process___iter__
    cls._process___contains__ = _process___contains__
    cls._process___len__ = _process___len__
    cls._process_clear = _process_clear
    cls._process_keys = _process_keys
    cls._process_values = _process_values
    cls._process_items = _process_items
    cls._process_pop = _process_pop

    return cls



"""Defines the CloseableQueue class and the Close exception class."""

from time import monotonic as time
import queue


class Closed(Exception):
    """Exception raised by CloseableQueue.put/get on a closed queue."""
    pass


class CloseableQueue(queue.Queue):
    """This class provides a means to permanently close a queue.

    Attempts to `put` to a closed queue will raise the `Closed` exception.

    Attempts to `get` from an *empty* closed queue will raise the same.

    Blocked `put`s and `get`s on a queue which is subsequently closed
      will also raise the `Closed` exception under the same circumstances.

    A queue can be closed either by calling its `close` method
      or by passing `last=True` to an invocation of `put`.

    If the latter is done, the entire operation is performed atomically;
      the close will only take place if the put succeeds.
    """
    def __init__(self, *args, **kwargs):
        queue.Queue.__init__(self, *args, **kwargs)
        assert not hasattr(self, '_closed')
        self._closed = False

    def close(self):
        """Close the queue.

        This will prevent further `put`s, and only allow `get`s
          until the contents are depleted.

        `put`s and `get`s which are prevented raise `Closed`.

        Calling `close` will also cause `Closed` exceptions to be raised
          in blocked `get`s or `put`s as though they had just been called.

        Normally it is only useful to call this method
          from a thread which is the sole producer or sole consumer.
        """
        self.mutex.acquire()
        try:
            if not self._closed:
                self._closed = True
                self.not_empty.notify_all()
                self.not_full.notify_all()
        finally:
            self.mutex.release()

    def closed(self):
        """True iff the queue is closed.
        Unreliable like `empty` and `full`."""
        # Probably not necessary to use a protected section here,
        #   but better safe than bug-ridden.
        # This also leaves the door open
        #   for more complex implementations of `_closed` as a property.
        self.mutex.acquire()
        n = self._closed
        self.mutex.release()
        return n

    def put(self, item, block=True, timeout=None, last=False):
        """Put an item into the queue.

        Works as does `Queue.Queue.put`, but with these differences:

        If the queue is closed, raises Closed.

        If `last` is True and the put succeeds,
          the queue will be atomically closed.

        Also raises `Closed` in the event that the queue is closed
          while the `put` is blocked.
        """
        self.not_full.acquire()
        try:
            if self.maxsize > 0:
                if not block:
                    if self._qsize() == self.maxsize and not self._closed:
                        raise queue.Full
                elif timeout is None:
                    while self._qsize() == self.maxsize and not self._closed:
                        self.not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a positive number")
                else:
                    endtime = time() + timeout
                    while self._qsize() == self.maxsize and not self._closed:
                        remaining = endtime - time()
                        if remaining <= 0.0:
                            raise queue.Full
                        self.not_full.wait(remaining)
            if self._closed:
                raise Closed
            self._put(item)
            self.unfinished_tasks += 1
            if last:
                self._closed = True
                self.not_empty.notify_all()
                self.not_full.notify_all()
            else:
                self.not_empty.notify()
        finally:
            self.not_full.release()

    def get(self, block=True, timeout=None):
        """Remove and return an item from the queue.

        Works as does `Queue.Queue.get` except that
          a `get` on a closed queue will raise `Closed`.

        Similarly, a blocked `get` will raise `Closed`
          if the queue is closed during the block.
        """
        self.not_empty.acquire()
        try:
            if not block:
                if not self._qsize() and not self._closed:
                    raise queue.Empty
            elif timeout is None:
                while not self._qsize() and not self._closed:
                    self.not_empty.wait()
            elif timeout < 0:
                raise ValueError("'timeout' must be a positive number")
            else:
                endtime = time() + timeout
                while not self._qsize() and not self._closed:
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise queue.Empty
                    self.not_empty.wait(remaining)
            if self._closed and not self._qsize():
                raise Closed
            item = self._get()
            self.not_full.notify()
            return item
        finally:
            self.not_empty.release()


def dequeue(q, getargs={}, on_empty='stop'):
    """Generates values from the queue `q`.

    This is a fairly flexible function which can also be meaningfully applied
      to non-closeable queues, by passing a `timeout` in `getargs`.

    The default arguments are suitable for generating an iteration
      from a queue which is closed after/as the last value is `put` to it.

    To avoid false StopIterations, if a timeout is used with a closeable queue,
      `on_empty` should be passed as `raise` or some sentinel value.

    `getargs` is a dict which will comprise the keyword arguments to `q.get`.

    `on_empty` determines the response to an `Empty` exception raised by `get`.
      If 'raise', the `Empty` exception will be raised.
      If 'stop', the `Empty` exception will be treated as a `Closed` exception.
      Otherwise, the value passed in `on_empty` will be yielded.

    Note that with the default value of `getargs`,
      `Empty` exceptions will not occur.
    """
    try:
        while True:
            yield q.get(**getargs)
    except Closed:
        raise StopIteration
    except queue.Empty:
        if on_empty == 'raise':
            raise
        elif on_empty == 'stop':
            raise StopIteration
        else:
            yield on_empty


def enqueue(it, q, putargs={}, join=False, close=True):
    """`put`s the successive values of the iterable `it` into `q`.

    The default values will close the queue after the final iterated value.

    `putargs` is a dict which will comprise the keyword arguments to `q.put`.

    If `close` is true,
      `q` must support the `close` method, i.e. be a CloseableQueue.
    This will have the effect of closing the queue after the end of iteration.

    If `join` is true, the queue is joined after the values are put,
      and after optionally being closed.
    """
    for value in iter(it):
        q.put(value, **putargs)
    if close:
        q.close()
    if join:
        q.join()


def EnqueueThread(it, q=None, name='enqueue', start=True, enqueue=enqueue,
                  **kwargs):
    """Starts a thread which enqueues the values of the iterable `it`.

    If a queue is not passed, a new one is created.

    A reference to the queue is stored as the property `q`
      of the returned thread.

    The thread will be started unless `start` is false.

    This is ideal for processing generators
      and other iterables which lazily process I/O-bound operations.

    Passing the return value of a `dequeue` call will have the effect
      of creating a processing thread which pulls from and pushes to
      thread-safe data structures.

    Additional keyword arguments are passed on to `enqueue`.
    """
    from threading import Thread
    if q is None:
        q = CloseableQueue()
    thread = Thread(name=name,
                    target=enqueue,
                    args=(it, q),
                    kwargs=kwargs)
    thread.q = q
    if start:
        thread.start()
    return thread



from __future__ import print_function
import hashlib
import sys
import os
import netifaces


def get_ip():
    for interface in netifaces.interfaces():
        addresses = netifaces.ifaddresses(interface)
        all_infos = addresses.setdefault(netifaces.AF_INET, [{'addr': ''}])
        for info in all_infos:
            if info['addr'] not in ['127.0.0.1', '']:
                return info['addr']
    return '127.0.0.1'


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def md5(file_name):
    if not os.path.isfile(file_name):
        return None
    else:
        m = hashlib.md5()
        file = open(file_name, 'rb')
        while True:
            data = file.read(4096)
            if not data:
                break
            m.update(data)
        file.close()

        return m.hexdigest()


def file_size(file_name):
    return (os.path.getsize(file_name) if os.path.isfile(file_name) else 0)



import threading
import time
import ctypes
import collections

from .CloseableQueue import CloseableQueue
from .CloseablePipe import CloseablePipe
from .utils import eprint


class OrderedDict(collections.OrderedDict):

    def __init__(self, **kwargs):
        collections.OrderedDict.__init__(self, **kwargs)

    def __setitem__(self, name, value):
        if isinstance(name, int):
            name = list(self.keys())[name]

        collections.OrderedDict.__setitem__(self, name, value)

    def __getitem__(self, name):
        if isinstance(name, int):
            name = list(self.keys())[name]

        return collections.OrderedDict.__getitem__(self, name)

    def __delitem__(self, name):
        if isinstance(name, int):
            name = list(self.keys())[name]

        collections.OrderedDict.__delitem__(self, name)


class QueueDict(dict):

    def __init__(self):
        dict.__init__(self)

    def __getitem__(self, name):
        if not dict.__contains__(self, name):
            new_queue = CloseableQueue()
            dict.__setitem__(self, name, new_queue)
        return dict.__getitem__(self, name)

    def __delitem__(self, name):
        if dict.__contains__(self, name):
            try:
                self[name].close()
            except BaseException:
                pass
            dict.__delitem__(self, name)

    def clear(self):
        for name in self:
            try:
                self[name].close()
            except BaseException:
                pass
        dict.clear(self)


class PipeDict(dict):

    def __init__(self):
        dict.__init__(self)

    def __getitem__(self, name):
        if not dict.__contains__(self, name):
            dict.__setitem__(self, name, CloseablePipe())
        return dict.__getitem__(self, name)

    def __delitem__(self, name):
        if dict.__contains__(self, name):
            try:
                self[name][0].close()
            except BaseException:
                pass

            try:
                self[name][1].close()
            except BaseException:
                pass

            dict.__delitem__(self, name)

    def clear(self):
        for name in self:
            try:
                self[name][0].close()
            except BaseException:
                pass

            try:
                self[name][1].close()
            except BaseException:
                pass

        dict.clear(self)


class Pipe:

    def __init__(self, node, name):
        self._node = node
        self._name = name

    def send(self, value):
        self._node._locals_pipe_send(self._name, value)

    def recv(self):
        return self._node._locals_pipe_recv(self._name)


class Pipes:

    def __init__(self, node):
        self._node = node

    def __getitem__(self, name):
        return Pipe(self._node, name)

    def __delitem__(self, name):
        self._node._locals_pipes_delitem(name)

    def __len__(self):
        return self._node._locals_pipes_len()

    def __contains__(self, name):
        return self._node._locals_pipes_contains(name)

    def __iter__(self):
        return self._node._locals_pipes_iter()

    def keys(self):
        return self._node._locals_pipes_keys()

    def values(self):
        values = []
        for key in self.keys():
            values.append(self[key])

        return values

    def items(self):
        items = []
        for key in self.keys():
            items.append((key, self[key]))

        return items

    def clear(self):
        self._node._locals_pipes_clear()


class Queue:

    def __init__(self, node, name, private=False):
        self._node = node
        self._name = name
        self._is_private = private

    def get(self, timeout=None, block=True):
        return self._node._locals_queue_get(
            self._name, timeout=timeout,
            block=block, is_private=self._is_private
        )

    def put(self, value, timeout=None, block=True):
        self._node._locals_queue_put(
            self._name, value,
            timeout=timeout, block=block,
            is_private=self._is_private
        )

    def qsize(self):
        return self._node._locals_queue_len(
            self._name, is_private=self._is_private
        )

    def __len__(self):
        return self._node._locals_queue_len(
            self._name, is_private=self._is_private
        )


class Queues:

    def __init__(self, node, private=False):
        self._node = node
        self._is_private = private

    def __getitem__(self, name):
        return Queue(self._node, name, private=self._is_private)

    def __delitem__(self, name):
        self._node._locals_queues_delitem(name, is_private=self._is_private)

    def __len__(self):
        return self._node._locals_queues_len(is_private=self._is_private)

    def __contains__(self, name):
        return self._node._locals_queues_contains(
            name, is_private=self._is_private
        )

    def keys(self):
        return self._node._locals_queues_keys(is_private=self._is_private)

    def values(self):
        values = []
        for key in self.keys():
            values.append(self[key])

        return values

    def items(self):
        items = []
        for key in self.keys():
            items.append((key, self[key]))
        return items

    def __iter__(self):
        return self._node._locals_queues_iter(is_private=self._is_private)

    def clear(self):
        self._node._locals_queues_clear(is_private=self._is_private)


class ServerPeer:

    class Queue:
        def __init__(self, node, name):
            self._node = node
            self._name = name

        def get(self, timeout=None, block=True):
            return self._node._server_queue_get(
                self._name, timeout=timeout, block=block
            )

        def put(self, value, timeout=None, block=True):
            self._node._server_queue_put(
                self._name, value, timeout=timeout, block=block
            )

        def qsize(self):
            return self._node._server_queue_len(self._name)

        def __len__(self):
            return self._node._server_queue_len(self._name)

    class Queues:
        def __init__(self, node):
            self._node = node

        def __getitem__(self, name):
            return ServerPeer.Queue(self._node, name)

        def __delitem__(self, name):
            self._node._server_queues_delitem(name)

        def __len__(self):
            return self._node._server_queues_len()

        def __contains__(self, name):
            return self._node._server_queues_contains(name)

        def keys(self):
            return self._node._server_queues_keys()

        def values(self):
            values = []
            for key in self.keys():
                values.append(self[key])
            return values

        def items(self):
            items = []
            for key in self.keys():
                items.append((key, self[key]))
            return items

        def __iter__(self):
            return self._node._server_queues_iter()

        def clear(self):
            self._node._server_queues_clear()

    def __init__(self, node):
        self._node = node
        self._queues = ServerPeer.Queues(node)

    def __setitem__(self, name, value):
        self._node._server_setitem(name, value)

    def __getitem__(self, name):
        return self._node._server_getitem(name)

    def __delitem__(self, name):
        return self._node._server_delitem(name)

    def __iter__(self):
        return self._node._server_iter()

    def __contains__(self, name):
        return self._node._server_contains(name)

    def __len__(self):
        return self._node._server_len()

    def clear(self):
        self._node._server_clear()

    def keys(self):
        return self._node._server_keys()

    def values(self):
        return self._node._server_values()

    def items(self):
        return self._node._server_items()

    def pop(self, name):
        return self._node._server_pop(name)

    def get(self, timeout=None, block=True):
        return self._node._server_get(timeout=timeout, block=block)

    def put(self, value, timeout=None, block=True):
        self._node._server_put(value, timeout=timeout, block=block)

    def qsize(self):
        return self._node._server_qsize()

    @property
    def queues(self):
        return self._queues

    @property
    def address(self):
        try:
            return self._node._connection.getpeername()
        except BaseException:
            return None


class AutoSendBuffer:

    def __init__(self, node, session_id, timeout=0.1, quiet=False):
        self._node = node
        self._session_id = session_id
        self._buffer_lock = threading.Lock()
        self._buffer = b''
        self._quiet = quiet
        self._timeout = timeout
        self._last_send_time = time.time()
        self._continue_send = True
        self._sending_thread = threading.Thread(target=self._timeout_send)
        self._sending_thread.start()

    def __del__(self):
        try:
            self.stop()
        except BaseException:
            pass

    def append(self, byte):
        with self._buffer_lock:
            self._buffer += byte

        if byte == b"\n" or len(self._buffer) >= 512:
            self.send()

    def send(self):
        with self._buffer_lock:
            if len(self._buffer) != 0:
                content = self._buffer.decode("utf-8")
                if not self._quiet:
                    print(content, end="", flush=True)
                self._node._respond_ok(
                    self._session_id, end=False, stdout=content
                )
                self._buffer = b''
                self._last_send_time = time.time()

    def stop(self):
        self._continue_send = False
        self._sending_thread.join()
        self.send()

    def _timeout_send(self):
        while self._continue_send:
            if time.time()-self._last_send_time >= self._timeout:
                self.send()
            time.sleep(1E-3)


class Thread(threading.Thread):

    def __init__(self, target, args=(), kwargs={}):
        threading.Thread.__init__(
            self, target=target,
            args=args, kwargs=kwargs
        )

    def __get_id(self):
        if hasattr(self, '_thread_id'):
            return self._thread_id

        for id, thread in threading._active.items():
            if thread is self:
                return id

    def kill(self):
        thread_id = self.__get_id()
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
            thread_id, ctypes.py_object(SystemExit)
        )
        if res > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
            print('Exception raise failure')


class Future:

    def __init__(self, node, session_id):
        self._node = node
        self._session_id = session_id
        self._cancelled = False
        self._done = None
        self._result = None
        self._monitor_done_thread = None
        self._done_callbacks = []

    def cancel(self):
        if self.done():
            return False

        self._node._stop_send = True
        self._node._send_signal(self._session_id, cancel=True)
        result_queues = self._node._result_queues_for_future
        self._result = result_queues[self._session_id].get()
        self._node._close_session(self._session_id)
        self._cancelled = True
        self._done = True
        return True

    def cancelled(self):
        return self._cancelled

    def done(self):
        if self._done is not None:
            return self._done
        else:
            result_queues = self._node._result_queues_for_future
            return (len(result_queues[self._session_id]) > 0)

    def result(self):
        if self._result is None:
            result_queues = self._node._result_queues_for_future
            self._result = result_queues[self._session_id].get()
            self._node._close_session(self._session_id)

        self._done = True
        if self._result["success"]:
            return self._result["return_value"]
        else:
            eprint(self._result["traceback"])
            raise self._result["exception"]

    def exception(self):
        if self._result is None:
            result_queues = self._node._result_queues_for_future
            self._result = result_queues[self._session_id].get()
            self._node._close_session(self._session_id)

        self._done = True
        return self._result["exception"]

    def _get_result(self):
        if self._result is None:
            result_queues = self._node._result_queues_for_future
            self._result = result_queues[self._session_id].get()
            self._node._close_session(self._session_id)

        for func in self._done_callbacks:
            func(self)

    def add_done_callback(self, func):
        self._done_callbacks.append(func)

        if self._monitor_done_thread is None:
            self._monitor_done_thread = threading.Thread(
                target=self._get_result
            )
            self._monitor_done_thread.start()

    def remove_done_callback(self, func):
        try:
            self._done_callbacks.remove(func)
        except BaseException:
            pass
